package v1;
import java.io.ObjectInputStream.GetField;
import java.util.Collections;
import java.util.HashSet;
import java.util.Set;
import java.util.TreeSet;

public class Profit {

	// The dynamic program used to find port values. States represent labels. 
	
	private static int maxWaitDays = 20; // Max wait days between each trade
	private static double EOHPerDayValue = 400000;  // EOH value per day
	
	public static double [] getProfit(Grid grid, int portNr, int timeToPort, int nrDays,int maxWaitDaysInPort, int capacity, int maxTradeHorizon, int currentTime, int totalTradePeriod) {
		double [] returnValues = new double [maxWaitDaysInPort];
		long startTime = System.currentTimeMillis();
		Set<State> states = new HashSet<State>();
		for (int k = 0; k < maxWaitDaysInPort; k++) { // Number of wait days before making the first trade
			State startState =  new State(grid.ports.get(portNr-1), "" + grid.ports.get(portNr-1).getNr() + "(" + (1 + timeToPort + k + 1) + ")" ,  1 + timeToPort + k + 1, (grid.ports.get(portNr-1).isBuyPort() ? -capacity * grid.ports.get(portNr-1).getPrice(1 + timeToPort + k + 1) : capacity * grid.ports.get(portNr-1).getPrice(1 + timeToPort + k + 1)) ,1,0);
			states.add(startState);
			HashSet<State> checkStates = new HashSet<State>(states);
			HashSet<State> newStates = new HashSet<State>(states);
			int prevStatesSize = 0;
			int counter = 0;
			while(states.size() != prevStatesSize){ // While new states(labels) are added
				newStates.clear();
				prevStatesSize = states.size();
				for (State state : checkStates) {
					boolean killState = false;
					for (Port port : grid.ports) {
						int [][] distanceGrid = GridValues.getGridDistances(grid.verSize, grid.horSize, grid.portCoords[state.getPort().getNr()-1][0], grid.portCoords[state.getPort().getNr()-1][1]); 
						int visitDay = state.getDay() + 1 +  distanceGrid[grid.portCoords[port.getNr()-1][0]][grid.portCoords[port.getNr()-1][1]];
						int [][] gridCosts = GridValues.getGridCosts(grid.verSize, grid.horSize, grid.portCoords[state.getPort().getNr()-1][0], grid.portCoords[state.getPort().getNr()-1][1]);
						if(visitDay <= nrDays) { // Visit day of the trade does not exceed total problem period.
							int waitDays = Math.min(nrDays-visitDay,maxWaitDays);
							for (int j = 0; j <= waitDays; j++) { // Wait days before making the trade
								TreeSet<State> tempStates = new TreeSet<State>();
								if (state.getPort().isBuyPort() && !port.isBuyPort()) { // If selling
									double profitFromTrade = capacity * port.getPrice(visitDay+j);
									double travelCost = gridCosts[grid.portCoords[port.getNr()-1][0]][grid.portCoords[port.getNr()-1][1]];
									double shareOfMonth = (((double)visitDay+j))/30; 
									double discountFactor = 1.0+(shareOfMonth*0.01);
					
									double discountedTravelCost = travelCost/discountFactor; 
									double discountedProfit = profitFromTrade/discountFactor;
									double EOHValue = 0;
									int EOHDays = nrDays - (visitDay+j);
									if(EOHDays<10 && ((totalTradePeriod - currentTime) > maxTradeHorizon)) {
										EOHValue = EOHDays*EOHPerDayValue;
									}
									State newState = new State(port, state.getRoute() + "-V" + j + "-" + port.getNr() +  "(" + (visitDay+j) + ")" + "[" + state.getProfit() + "]", visitDay + j, state.getProfit() - discountedTravelCost + discountedProfit + EOHValue, state.getNrVisits()+1,state.getNrDaysWaited()+j);
									if(!states.contains(newState)) {
										killState = true;
										tempStates.add(newState);
										tempStates.addAll(states);
										states = new HashSet<State>(tempStates);
									}
								}
								else if(!state.getPort().isBuyPort() && port.isBuyPort()){ // If buying
									double profitFromTrade = capacity * port.getPrice(visitDay+j);
									double travelCost = gridCosts[grid.portCoords[port.getNr()-1][0]][grid.portCoords[port.getNr()-1][1]];
									double shareOfMonth = (((double)visitDay+j))/30; 
									double discountFactor = 1.0+(shareOfMonth*0.01);
									double discountedTravelCost = travelCost/discountFactor;
									double discountedProfit = profitFromTrade/discountFactor;
									double EOHValue = 0;
									int EOHDays = nrDays - (visitDay+j);
									if(EOHDays<10 && ((totalTradePeriod - currentTime) > maxTradeHorizon)){
										EOHValue = EOHDays*EOHPerDayValue;
									}
									if(EOHDays<5 && ((totalTradePeriod - currentTime) > maxTradeHorizon)) {
										EOHValue += discountedProfit*1.2; // Buy value compensation
									}
									State newState = new State(port, state.getRoute() + "-V" + j  + "-" + port.getNr() + "(" + (visitDay+j) + ")" + "[" + state.getProfit() + "]", visitDay + j, state.getProfit() - discountedTravelCost - discountedProfit + EOHValue, state.getNrVisits()+1,state.getNrDaysWaited()+j);
									if(!states.contains(newState) && ((counter>2 && newState.getProfit()>0) || counter <=2)) {
										killState = true;
										tempStates.add(newState);
										tempStates.addAll(states);
										states = new HashSet<State>(tempStates);
									}
								}

							}

						}
					}
					if (killState && state!=Collections.max(states))
						states.remove(state);
				}	
				checkStates = new HashSet<State>(states);
				counter++;
			}
			State maxState = Collections.max(states);
			double EOH = 0;
			returnValues[k] =  maxState.getProfit() + EOH; 
			states.clear();
		}
		return returnValues;
	}
}
