package v1;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.io.ObjectInputStream.GetField;
import java.text.DecimalFormat;
import java.util.Arrays;

// Main file. MIP = true if only finding port values for MIP. Heuristic run if MIP = false. 

public class newMain {
	
	static boolean MIP = false; // True if calculating port values for the MIP model
	
	static int nrScenarios = 5;
	static int maxTradeHorizon = 60; //Maximum number of days considered by the DP
	static int nrRunDays = 200; // Total problem period 
	static boolean decHor = true; //True if using decreasing length of the rolling horizon
	static boolean onlyFewer = true; //True if only including ports that get closer in move calculations 
	static boolean toBuy = true; // True if the vessel starts empty
	static boolean stochastic = true; // True if doing a stochastic run of the heuristic
	static int verPos = 3; // Vertical start position
	static int horPos = 12;  // Horizontal start position
	static int maxWaitDaysInPort = 7; // Max wait days in port for the DP
	static int capacity = 3000000;  // Vessel capacity
	static int verSize = 13; // Vertical grid size
	static int horSize = 18; // Horizontal grid size
	static BufferedReader reader;
	static int currentTime = 1;
	static int nrPorts = 7; 
	
	public static void main(String[] args) {
		int tradeHorizon = Math.min(maxTradeHorizon, nrRunDays-currentTime); // Trade horizon considered by the DP
		int decreasingHor = tradeHorizon;
		BufferedReader input;
		if(MIP){
			try {				
				input = new BufferedReader(new FileReader("MIP.txt"));  
				String value = input.readLine(); 
				verPos = Integer.parseInt(value);
				value = input.readLine();
				horPos = Integer.parseInt(value);
				value = input.readLine();
				toBuy = Boolean.parseBoolean(value);
				value = input.readLine();
				currentTime = Integer.parseInt(value);
				if(decHor){
				value = input.readLine();
				decreasingHor = Integer.parseInt(value);
				}
			} catch (FileNotFoundException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}



		}	
		long startTime = System.currentTimeMillis();
		Grid grid = new Grid(150,nrPorts, verSize, horSize);
		double profit = 0;
		String route = verPos + "," + horPos;
		String profitString = "0";
	

		System.out.print("Starting position: " + verPos + ", " + horPos);
		System.out.print(stochastic ? ". Stochastic run" : ". Deterministic run");
		System.out.print(toBuy ? ". To buy. " : ". To sell. ");
		System.out.print(nrScenarios + " scenarios.");
		System.out.println(" Current time: " + currentTime);
		System.out.print("Rolling horizon length: ");
		System.out.println(decHor ? decreasingHor : tradeHorizon);
		while(currentTime < nrRunDays){
			double [][][] prices = ReadCSV.Prices(nrScenarios, nrPorts, maxTradeHorizon, currentTime, nrRunDays);
			//					for (int i = 0; i < prices.length; i++) {
			//						for (int j = 0; j < prices[0].length; j++) {
			//							for (int j2 = 0; j2 < prices[0][0].length; j2++) {
			//								System.out.print(prices[i][j][j2] + "  ");
			//							}
			//							System.out.println();
			//						}
			//						System.out.println("SDFSDFASDF");
			//					}
			if(MIP){				// In a MIP run we only need to calculate port vales. This is done in getValueGrid
				if(decHor){
					getValueGrid(grid,toBuy,stochastic, verPos, horPos, prices, decreasingHor);
				}
				else {
					getValueGrid(grid,toBuy,stochastic, verPos, horPos, prices, tradeHorizon);
				}
				System.exit(0);
			}
	
			// The following code lines belong to the heuristic run
			
			int daysBeforeTrade; 
			int portNr = inPort(grid, verPos, horPos); 
			boolean tradeDone = false;
			if(portNr != 99 && (grid.ports.get(portNr-1).isBuyPort() == toBuy)) {    //Returns 99 if not located in a port
				while(!tradeDone){ 
					System.out.println("I port! portNr: " + portNr);
					if(decHor){ 
						daysBeforeTrade = findWaitDaysBeforeTrade(grid, portNr, stochastic, prices, decreasingHor); // The DP calculates when it is optimal to make the trade
					}
					else {
						daysBeforeTrade = findWaitDaysBeforeTrade(grid, portNr, stochastic, prices, tradeHorizon);
					}
					System.out.println("Wait days before trade: " + daysBeforeTrade);
					System.out.println("Current time: " + currentTime);
					if(daysBeforeTrade>0){ // Waits until the next day
						currentTime++;
						decreasingHor--;
						tradeHorizon = Math.min(maxTradeHorizon, nrRunDays-currentTime);
						profitString += profit + ",";
						profit -= 100000;
						prices = ReadCSV.Prices(nrScenarios, nrPorts, maxTradeHorizon, currentTime, nrRunDays);
					}
					else { // Makes the trade 
						tradeDone = true;
						decreasingHor = tradeHorizon;
						currentTime++;
						tradeHorizon = Math.min(maxTradeHorizon, nrRunDays-currentTime);
						prices = ReadCSV.Prices(nrScenarios, nrPorts, maxTradeHorizon, currentTime, nrRunDays);
						route += "(" + currentTime + ")";
						route += (toBuy ? "(B)" : "(S)");
						profit += capacity*(toBuy ? -prices[0][portNr-1][0]:prices[0][portNr-1][0]);
						toBuy = !toBuy;
						System.out.println("Profit after trade: " + profit);
						profitString += profit + ",";
					}


				}
			}
			
			int [] newPos = new int [2];
			
			if(decHor){
				newPos = findNewPos(grid, toBuy, verPos, horPos, stochastic, prices, decreasingHor); // Finds an updated position for the vessel
			}
			else {
				newPos = findNewPos(grid, toBuy, verPos, horPos, stochastic, prices, tradeHorizon);
			}
			
			if((newPos[0] != verPos) && (newPos[1] != horPos) )
				profit -= 240000;
			else if(newPos[0] != verPos || newPos[1] != horPos)
				profit -= 200000;
			else
				profit -= 100000;
			verPos = newPos[0];
			horPos = newPos[1];
			currentTime++;
			decreasingHor--;
			tradeHorizon = Math.min(maxTradeHorizon, nrRunDays-currentTime);
			profitString += profit + ",";
			route += "-" + verPos + "," + horPos + "(" + currentTime + ")"; 
			System.out.print("Current time: " + currentTime +". Trade horizon: " + tradeHorizon);
			System.out.print(toBuy ? ". To buy. " :". To sell. " );
			System.out.print("Profit: " + profit);
			System.out.println(". Next position: " + verPos +", " + horPos + ".");
			System.out.println(route);
			System.out.println(profitString);
		}
		
		long endTime   = System.currentTimeMillis();
		long totalTime = (endTime - startTime);
		System.out.println("Total time spent: " + totalTime + " ms.");
		try {
			PrintWriter out = new PrintWriter("runResult.txt");	
			out.println("Route: ");
			out.println(route);
			out.println("Profit: ");
			out.println(profitString);
			out.println("Final profit: " + profit);
			out.println("Time spent: " + totalTime);
			
			out.close();
		} catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
	}





	public static int inPort(Grid grid, int verPos, int horPos) {
		for (int i = 0; i < grid.portCoords.length; i++) {
			if(grid.portCoords[i][0] == verPos && grid.portCoords[i][1] == horPos)
				return i+1;
		}
		return 99;
	}

	public static int [] findNewPos(Grid grid, boolean toBuy, int verPos, int horPos, boolean stochastic, double [][][] prices, int tradeHorizon){
		System.out.println("Find new pos");
		int [] newPosStoch = {verPos, horPos};
		double [][][] valueGrid = getValueGrid(grid,toBuy,stochastic, verPos, horPos, prices, tradeHorizon);

		
		for (int i = 1; i < verSize+1; i++) {
			for (int j = 1; j < horSize+1; j++) {
				if(((verPos+1) == i) && ((horPos+1) == j)){
					System.out.print("!!");
				}
				else{
					System.out.print("  ");
				}

				System.out.format("%.0f", valueGrid[0][i][j]);
				boolean check = false;
				for (int j2 = 0; j2 < grid.portCoords.length; j2++) {
					if(grid.portCoords[j2][0]+1 == i && grid.portCoords[j2][1]+1 == j){
						check = true;
						if(grid.ports.get(j2).isBuyPort())
							System.out.print("(B" + (j2+1)+")");
						else
							System.out.print("(S"+ (j2+1)+ ")");
					}	
				}
				if(!check)
					System.out.print("    ");
				check=false; 
				if(valueGrid[0][i][j]== 0)
					System.out.print("       ");
				else if(valueGrid[0][i][j]< 0)
					System.out.print("      ");
				else
					System.out.print("");
				System.out.print("");
			}
			System.out.println();
		}

		
		

		double currentMaxValueStoch = valueGrid[0][verPos+1][horPos+1];

		// Calculates the value of each potential move

		//-1,-1
		if(valueGrid[0][verPos][horPos] == currentMaxValueStoch){
			double highestPortValueInCurrentPoint = 0;
			double highestPortValueInPotentialPoint = 0;
			for (int i = 1; i < valueGrid.length; i++) {
				highestPortValueInCurrentPoint = Math.max(highestPortValueInCurrentPoint, valueGrid[i][newPosStoch[0]+1][newPosStoch[1]+1]); 
				highestPortValueInPotentialPoint = Math.max(highestPortValueInPotentialPoint,valueGrid[i][verPos][horPos]);
			}
			if(highestPortValueInCurrentPoint<highestPortValueInPotentialPoint){
				newPosStoch[0] = verPos-1;
				newPosStoch[1] = horPos-1;
			}
		}

		if(valueGrid[0][verPos][horPos] > currentMaxValueStoch){
			//	System.out.println("-1,-1");
			currentMaxValueStoch = valueGrid[0][verPos][horPos];
			newPosStoch[0] = verPos-1;
			newPosStoch[1] = horPos-1;
		}

		if(valueGrid[0][verPos][horPos+1] == currentMaxValueStoch){
			double highestPortValueInCurrentPoint = 0;
			double highestPortValueInPotentialPoint = 0;
			for (int i = 1; i < valueGrid.length; i++) {
				highestPortValueInCurrentPoint = Math.max(highestPortValueInCurrentPoint, valueGrid[i][newPosStoch[0]+1][newPosStoch[1]+1]); 
				highestPortValueInPotentialPoint = Math.max(highestPortValueInPotentialPoint,valueGrid[i][verPos][horPos+1]);
			}
			if(highestPortValueInCurrentPoint<highestPortValueInPotentialPoint){
				newPosStoch[0] = verPos-1;
				newPosStoch[1] = horPos;
			}
		}

		if (valueGrid[0][verPos][horPos+1] > currentMaxValueStoch) {
			//	System.out.println("-1,0");
			currentMaxValueStoch = valueGrid[0][verPos][horPos+1];
			newPosStoch[0] = verPos-1;
			newPosStoch[1] = horPos;
		}

		if(valueGrid[0][verPos][horPos+2] == currentMaxValueStoch){
			double highestPortValueInCurrentPoint = 0;
			double highestPortValueInPotentialPoint = 0;
			for (int i = 1; i < valueGrid.length; i++) {
				highestPortValueInCurrentPoint = Math.max(highestPortValueInCurrentPoint, valueGrid[i][newPosStoch[0]+1][newPosStoch[1]+1]); 
				highestPortValueInPotentialPoint = Math.max(highestPortValueInPotentialPoint,valueGrid[i][verPos][horPos+2]);
			}
			if(highestPortValueInCurrentPoint<highestPortValueInPotentialPoint){
				newPosStoch[0] = verPos-1;
				newPosStoch[1] = horPos+1;		
			}
		}


		if (valueGrid[0][verPos][horPos+2] > currentMaxValueStoch) {
			//			System.out.println("-1,+1");
			currentMaxValueStoch = valueGrid[0][verPos][horPos+2];
			newPosStoch[0] = verPos-1;
			newPosStoch[1] = horPos+1;
		}

		if(valueGrid[0][verPos+1][horPos] == currentMaxValueStoch){
			double highestPortValueInCurrentPoint = 0;
			double highestPortValueInPotentialPoint = 0;
			for (int i = 1; i < valueGrid.length; i++) {
				highestPortValueInCurrentPoint = Math.max(highestPortValueInCurrentPoint, valueGrid[i][newPosStoch[0]+1][newPosStoch[1]+1]); 
				highestPortValueInPotentialPoint = Math.max(highestPortValueInPotentialPoint,valueGrid[i][verPos+1][horPos]);
			}
			if(highestPortValueInCurrentPoint<highestPortValueInPotentialPoint){
				newPosStoch[0] = verPos;
				newPosStoch[1] = horPos-1;	
			}
		}
		if (valueGrid[0][verPos+1][horPos] > currentMaxValueStoch) {
			//			System.out.println("1,-1");
			currentMaxValueStoch = valueGrid[0][verPos+1][horPos];
			newPosStoch[0] = verPos;
			newPosStoch[1] = horPos-1;
		}

		if(valueGrid[0][verPos+1][horPos+2] == currentMaxValueStoch){
			double highestPortValueInCurrentPoint = 0;
			double highestPortValueInPotentialPoint = 0;
			for (int i = 1; i < valueGrid.length; i++) {
				highestPortValueInCurrentPoint = Math.max(highestPortValueInCurrentPoint, valueGrid[i][newPosStoch[0]+1][newPosStoch[1]+1]); 
				highestPortValueInPotentialPoint = Math.max(highestPortValueInPotentialPoint,valueGrid[i][verPos+1][horPos+2]);
			}
			if(highestPortValueInCurrentPoint<highestPortValueInPotentialPoint){
				newPosStoch[0] = verPos;
				newPosStoch[1] = horPos+1;				
			}
		}
		if (valueGrid[0][verPos+1][horPos+2] > currentMaxValueStoch) {
			//			System.out.println("0,1");
			currentMaxValueStoch = valueGrid[0][verPos+1][horPos+2];
			newPosStoch[0] = verPos;
			newPosStoch[1] = horPos+1;
		}

		if(valueGrid[0][verPos+2][horPos] == currentMaxValueStoch){
			double highestPortValueInCurrentPoint = 0;
			double highestPortValueInPotentialPoint = 0;
			for (int i = 1; i < valueGrid.length; i++) {
				highestPortValueInCurrentPoint = Math.max(highestPortValueInCurrentPoint, valueGrid[i][newPosStoch[0]+1][newPosStoch[1]+1]); 
				highestPortValueInPotentialPoint = Math.max(highestPortValueInPotentialPoint,valueGrid[i][verPos+2][horPos]);
			}
			if(highestPortValueInCurrentPoint<highestPortValueInPotentialPoint){
				newPosStoch[0] = verPos+1;
				newPosStoch[1] = horPos-1;				
			}
		}

		if (valueGrid[0][verPos+2][horPos] > currentMaxValueStoch) {
			//			System.out.println("1,-1");
			currentMaxValueStoch = valueGrid[0][verPos+2][horPos];
			newPosStoch[0] = verPos+1;
			newPosStoch[1] = horPos-1;
		}

		if(valueGrid[0][verPos+2][horPos+1] == currentMaxValueStoch){
			double highestPortValueInCurrentPoint = 0;
			double highestPortValueInPotentialPoint = 0;
			for (int i = 1; i < valueGrid.length; i++) {
				highestPortValueInCurrentPoint = Math.max(highestPortValueInCurrentPoint, valueGrid[i][newPosStoch[0]+1][newPosStoch[1]+1]); 
				highestPortValueInPotentialPoint = Math.max(highestPortValueInPotentialPoint,valueGrid[i][verPos+2][horPos+1]);				
			}
			if(highestPortValueInCurrentPoint<highestPortValueInPotentialPoint){
				newPosStoch[0] = verPos+1;
				newPosStoch[1] = horPos;
			}
		}

		if (valueGrid[0][verPos+2][horPos+1] > currentMaxValueStoch) {
			//			System.out.println("1,0");
			currentMaxValueStoch = valueGrid[0][verPos+2][horPos+1];
			newPosStoch[0] = verPos+1;
			newPosStoch[1] = horPos;
		}

		if(valueGrid[0][verPos+2][horPos+2] == currentMaxValueStoch){
			double highestPortValueInCurrentPoint = 0;
			double highestPortValueInPotentialPoint = 0;
			for (int i = 1; i < valueGrid.length; i++) {
				highestPortValueInCurrentPoint = Math.max(highestPortValueInCurrentPoint, valueGrid[i][newPosStoch[0]+1][newPosStoch[1]+1]); 
				highestPortValueInPotentialPoint = Math.max(highestPortValueInPotentialPoint,valueGrid[i][verPos+2][horPos+2]);
			}
			if(highestPortValueInCurrentPoint<highestPortValueInPotentialPoint){
				newPosStoch[0] = verPos+1;
				newPosStoch[1] = horPos+1;
			}

		}

		if (valueGrid[0][verPos+2][horPos+2] > currentMaxValueStoch) {
			//			System.out.println("1,1");
			currentMaxValueStoch = valueGrid[0][verPos+2][horPos+2];
			newPosStoch[0] = verPos+1;
			newPosStoch[1] = horPos+1;
		}

		
		
		if(currentMaxValueStoch<=0){ //Ensures that the vessel stays put if no profitable options exist
			newPosStoch[0] = verPos;
			newPosStoch[1] = horPos;
		}

			return newPosStoch;
	}


	public static double [][][] getValueGrid(Grid grid, boolean toBuy, boolean stochastic, int verPos, int horPos, double [][][] prices, int tradeHorizon){
		double [][][] stochGrids = new double [toBuy ? grid.getNrOfBuyPorts() + 1 : grid.ports.size()-grid.getNrOfBuyPorts() + 1][verSize+2][horSize + 2]; //The first layer contains the total, the following for each of the ports 
		int nrBuyPorts = grid.getNrOfBuyPorts(); 
		int lowCheckIndex;
		int highCheckIndex;
		if(toBuy) { // Finds the relevant ports to be checked
			lowCheckIndex = 0;
			highCheckIndex = nrBuyPorts;
		}
		else {
			lowCheckIndex = nrBuyPorts;
			highCheckIndex = grid.ports.size();
		}

		double [][][] portValues = new double [nrScenarios][nrPorts][20];


		for (int i = 0; i < nrScenarios; i++) { 

			//			System.out.println("Scenario: " + (i+1));
			int counter = 0; 
			for (int j = 0; j < grid.ports.size(); j++) {
				grid.ports.get(j).setPrices(prices[i][j]);				
			}
			for (int k = lowCheckIndex; k < highCheckIndex; k++) { // Goes through each of the potential trade ports;
				counter++; //The total grid goes in the first level, so we need the first port specific grid in nr 1. 
				int [][] distanceGrid = GridValues.getGridDistances(grid.verSize, grid.horSize, grid.portCoords[k][0], grid.portCoords[k][1]); //Calculates distances between grid points
				int timeToPort = distanceGrid[verPos+1][horPos+1];
				System.out.println("Port: " + k +". Time to port: " + timeToPort );
				//				System.out.println("To buy: " + toBuy + ". Check nr: " + k +". Time to port: "  + timeToPort);
				double[] profits = Profit.getProfit(grid, k+1, timeToPort, tradeHorizon,7,capacity, maxTradeHorizon, currentTime, nrRunDays); // DP calculation of port values
				
				if(MIP){
					for (int j = timeToPort; j < profits.length+timeToPort; j++) {
						if(j<20)
							portValues[i][k][j] = profits[j-timeToPort];
					}
				}
				int maxDay = 0;
				for (int j = 1; j < profits.length; j++) {
					if (profits[j]>profits[maxDay])
						maxDay = j;
				}
				System.out.println("Scenario: " + i + ". Port: " + k + ". Waitdays: " + maxDay + ".");
				//				System.out.println();
				//				//		System.out.println("Port: " + (k+1) + ". Scenario: " + i + ". Optimal nr. of wait days: " + initDays + ". Max profit: " + maxValue);
				double [][] gridValues = GridValues.getGridValues(grid.verSize,grid.horSize,grid.portCoords[k][0],grid.portCoords[k][1], verPos, horPos, profits, onlyFewer);


				for (int j = 0; j < verSize +2 ; j++) {
					for (int j2 = 0; j2 < horSize + 2 ; j2++) {
						stochGrids[counter][j][j2] = gridValues[j][j2];
					}
				}


			}

			
			// Iterates through each of the ports and finds the port that contributes with the highest value for this scenario. The value of this scenario is added to the main grid

			for (int j = 0; j < verSize +2 ; j++) {
				for (int j2 = 0; j2 < horSize + 2 ; j2++) {
					double currMaxValue = stochGrids[1][j][j2];
					for (int k = 2; k < stochGrids.length; k++) {
						if(stochGrids[k][j][j2] > currMaxValue)
							currMaxValue = stochGrids[k][j][j2]; 
					}
					stochGrids[0][j][j2] += currMaxValue; 
				}
			}

		}



		if(MIP){
			try {
				PrintWriter out = new PrintWriter("verdiData.txt");	 // Writes out the port values so they can be read by the Mosel code
				out.printf("SPrice: [");
				for (int j = 0; j < portValues.length; j++) {
					for (int j2 = 0; j2 < portValues[0].length; j2++) {
						for (int k = 0; k < portValues[0][0].length; k++) {
							out.printf("%f", portValues[j][j2][k]);
							out.print(" ");
						}

					}

				}
				out.printf("]");
				out.close();
			} catch (FileNotFoundException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}	
		}
		if(MIP)
			System.exit(0);
		
		return stochGrids;
	}

	public static int findWaitDaysBeforeTrade(Grid grid, int portNr, boolean stoch, double[][][] prices, int tradeHorizon) { 
		int returnDaysStoch = 0;
		int returnDaysDet = 0;
		double [] detProfits = new double[maxWaitDaysInPort];
		double [] stochProfits = new double[maxWaitDaysInPort];
		for (int i = 0; i < +1; i++) {
			for (int j = 0; j < grid.ports.size(); j++) {
				grid.ports.get(j).setPrices(prices[i][j]);
			}

			double[] profits = Profit.getProfit(grid, portNr,0, tradeHorizon, maxWaitDaysInPort, capacity, maxTradeHorizon, currentTime, nrRunDays); // Runs the DP to get values for each potential trade day
			for (int j = 0; j < maxWaitDaysInPort; j++) {
				stochProfits[j] += profits[j];
				detProfits[j] += profits[j]; 
			}
		}

		double maxProfitStoch = stochProfits[0]; 
		for (int i = 0; i < maxWaitDaysInPort; i++) {
			if(stochProfits[i]> maxProfitStoch){
				maxProfitStoch = stochProfits[i];
				returnDaysStoch = i;
			}	
		}

		returnDaysStoch = Math.min(returnDaysStoch, tradeHorizon-1);
		return returnDaysStoch;
	}

	public static double [][] findStochGrid(){
		double [][] stochGrid = new double [verSize+2][horSize+2];
		return stochGrid;
	}


}